﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;           ///< To use Toradex C Win32 DLLs.
using HANDLE = System.IntPtr;

namespace vcsharp_interrupt_demo
{
    public partial class Form1 : Form
    {
        [DllImport("GPIOLib.dll", CharSet = CharSet.Auto)]      ///< Importing DLL.
        public static extern void InitGPIOLib();                ///< External function declaration.

        [DllImport("GPIOLib.dll", CharSet = CharSet.Auto)]                              
        public static extern bool SetPinAltFn(int pinNum, int altFn, bool dirOut);

        [DllImport("GPIOLib.dll", CharSet = CharSet.Auto)]
        public static extern bool SetPinLevel(int gpioNum, int val);

        [DllImport("GPIOLib.dll", CharSet = CharSet.Auto)]
        public static extern bool GetPinLevel(int pin);

        [DllImport("GPIOLib.dll", CharSet = CharSet.Auto)]
        public static extern bool GetGPIOFromPin(int pin, bool extensionConnector, ref int pInst);

        [DllImport("GPIOLib.dll", CharSet = CharSet.Auto)]
        public static extern void DeInitGPIOLib();

        [DllImport("IntLib.dll", CharSet = CharSet.Auto)]
        public static extern int GetGPIOIrq(int dwGpio);

        [DllImport("IntLib.dll", CharSet = CharSet.Auto)]
        public static extern bool SetGPIOIrqEdge(int dwGpio, int dwEdge);

        [DllImport("IntLib.dll", CharSet = CharSet.Auto)]
        public static extern int RequestSysInterrupt(int dwIrq);

        [DllImport("IntLib.dll", CharSet = CharSet.Auto)]
        public static extern bool InterruptInitializeCompat(int idInt, HANDLE hEvent, IntPtr pvData, int cbData);

        [DllImport("IntLib.dll", CharSet = CharSet.Auto)]
        public static extern void InterruptDoneCompat(int idInt);

        [DllImport("IntLib.dll", CharSet = CharSet.Auto)]
        public static extern void InterruptDisableCompat(int idInt);

        [DllImport("IntLib.dll", CharSet = CharSet.Auto)]
        public static extern bool ReleaseSysIntr(int dwSysIntr);

        [DllImport("Coredll.dll", CharSet = CharSet.Auto)]
        public static extern HANDLE CreateEvent(IntPtr lpEventAttributes, bool bManualReset, bool bInitialState, string lpName);

        [DllImport("Coredll.dll", CharSet = CharSet.Auto)]
        public static extern int WaitForSingleObject(HANDLE hHandle, int dwMilliseconds);
        
        [DllImport("Coredll.dll", CharSet = CharSet.Auto)]
        public static extern bool CloseHandle(HANDLE hObject);

        private const int GPIO_EDGE_RISING = 1;
        private const int GPIO_EDGE_FALLING = 2;
        private const int GPIO_EDGE_BOTH = 3;
        private const int GPIO = -1;
        private const bool DIR_IN = false;
        private const bool DIR_OUT = true;

        private const int WAIT_OBJECT_0 = 0;
        private const int WAIT_TIMEOUT = 0x102;
        
        private const int INTERRUPT_PIN = 101;      ///< SODIMM pin 101 as GPIO input to capture the interrupt.
        private const int OUTPUT_PIN = 135;         ///< SODIMM pin 135 as GPIO output to blink the LED.

        private int interruptEdge = GPIO_EDGE_RISING;
        private int interruptCount = 0;
        private int systemInterrupt = 0;
        private bool systemInterruptGenerated = false;
        private bool initialized = false;
        private bool edgeChanged = false;
        private int gpioNumber = 0;
        private HANDLE hEvent = HANDLE.Zero;
        private bool outputPinLogic = false;
        
        private void InterruptConfig()
        {
            int irq = 0;
            int gpio = 0;
            bool retVal = false;
            SetPinAltFn(INTERRUPT_PIN, GPIO, DIR_IN);   ///< Sets INTERRUPT_PIN as GPIO input to capture the interrupt.
            SetPinAltFn(OUTPUT_PIN, GPIO, DIR_OUT);     ///< Sets OUTPUT_PIN as GPIO output to blink the LED.
            SetPinLevel(OUTPUT_PIN, 0);                 ///< Sets OUTPUT_PIN level to LOW.
            GetGPIOFromPin(INTERRUPT_PIN, false, ref gpio);     ///< Gets GPIO number for SODIMM pin 101.
            gpioNumber = gpio & (0xFF);                 ///< Gets first instance.
            irq = GetGPIOIrq(gpioNumber);               ///< Requests interrupt (IRQ) on GPIO number.
            if (irq == 0)
            {
                MessageBox.Show("Failed to get IRQ.", "Error");
                return;
            }
            retVal = SetGPIOIrqEdge(gpioNumber, interruptEdge);
            if (retVal == false)
            {
                MessageBox.Show("Failed to set interrupt edge.", "Error");
                return;
            }
            systemInterrupt = RequestSysInterrupt(irq); ///< Gets the system interrupt number for corresponding irq.
            if (systemInterrupt == 0)
            {
                MessageBox.Show("Failed to get system interrupt number.", "Error");
                return;
            }
            systemInterruptGenerated = true;
            hEvent = CreateEvent(IntPtr.Zero, false, false, null);
            if (hEvent == HANDLE.Zero)
            {
                MessageBox.Show("Failed to generate event.", "Error");
                return;
            }
            /// Initializes the interrupt.
            retVal = InterruptInitializeCompat(systemInterrupt, hEvent, IntPtr.Zero, 0);
            if (retVal == false)
            {
                MessageBox.Show("Failed to initialize the interrupt.", "Error");
                return;
            }
            initialized = true;
            edgeSelectComboBox.Enabled = true;
            startButton.Enabled = true;
        }

        private bool RestartInterrupt()
        {
            bool retVal = false;
            if (initialized == true)
            {
                initialized = false;
                /// Disables system interrupt.
                InterruptDisableCompat(systemInterrupt);
                retVal = SetGPIOIrqEdge(gpioNumber, interruptEdge);
                if (retVal == false)
                {
                    return false;
                }
                /// Initializes the interrupt.
                retVal = InterruptInitializeCompat(systemInterrupt, hEvent, IntPtr.Zero, 0);
                initialized = true;
            }
            return retVal;
        }

        public Form1()
        {
            InitializeComponent();
            /// Initializes the GPIO library.
            InitGPIOLib();
            edgeSelectComboBox.SelectedIndex = 0;
            /// Configures and initializes the interrupt.
            InterruptConfig();
        }

        private void startButton_Click(object sender, EventArgs e)
        {
            bool retVal = false;
            if (initialized == true)                    ///< Starts only if initialization is done.
            {
                if (startButton.Text == "Start")
                {
                    edgeSelectComboBox.Enabled = false;
                    if (edgeChanged == true)
                    {
                        edgeChanged = false;
                        retVal = RestartInterrupt();    ///< Restarts interrupt to change the edge.
                        if (retVal == false)
                        {
                            textBox1.Text = "Failed.";
                            startButton.Enabled = false;
                            return;
                        }
                    }
                    startButton.Text = "Stop";
                    scanTimer.Enabled = true;
                }
                else
                {
                    scanTimer.Enabled = false;
                    System.Threading.Thread.Sleep(10);
                    edgeSelectComboBox.Enabled = true;
                    startButton.Text = "Start";
                }
            }
        }

        private void clearButton_Click(object sender, EventArgs e)
        {
            interruptCount = 0;
            textBox1.Text = "0";
        }

        private void edgeSelectComboBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            interruptEdge = edgeSelectComboBox.SelectedIndex + 1;
            edgeChanged = true;
        }

        private void scanTimer_Tick(object sender, EventArgs e)
        {
            int retVal = 0;
            retVal = WaitForSingleObject(hEvent, 10);   ///< Waits for event associated with interrupt.
            if (retVal == WAIT_OBJECT_0)
            {
                interruptCount++;
                textBox1.Text = interruptCount.ToString();
                if (outputPinLogic == false)
                {
                    SetPinLevel(OUTPUT_PIN, 1);
                    outputPinLogic = true;
                }
                else
                {
                    SetPinLevel(OUTPUT_PIN, 0);
                    outputPinLogic = false;
                }
                /// Clears the interrupt generated.
                InterruptDoneCompat(systemInterrupt);
            }
            else if (retVal != WAIT_TIMEOUT)
            {
                textBox1.Text = "Failed.";
                startButton_Click(sender, e);
            }
        }

        private void Form1_Closing(object sender, CancelEventArgs e)
        {
            scanTimer.Enabled = false;
            System.Threading.Thread.Sleep(10);
            if (initialized == true)
            {
                /// Disables system interrupt.
                InterruptDisableCompat(systemInterrupt);
            }
            if (hEvent != HANDLE.Zero)
            {
                CloseHandle(hEvent);
            }
            if (systemInterruptGenerated == true)
            {
                /// Releases all the resources related to system interrupt.
                ReleaseSysIntr(systemInterrupt);
            }
            /// De-initializes the GPIO library.
            DeInitGPIOLib();
        }
    }
}